from numpy import setdiff1d, intersect1d
from scipy.sparse import issparse
from os.path import isfile
from torch.nn.functional import cross_entropy
from torch.cuda import is_available
from torch.nn import MSELoss
from torch import load as torch_load, save,tensor

from .Utils import build_dir
from .Preprocessing import preprocess
from .Data_Infrastructure.DataLoader_Constructor import build_dataloaders
from .Network.Model import scMMT_Model
from .Network.Losses import no_loss,mse_loss

from sklearn.utils.class_weight import compute_class_weight
import torch
import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.decomposition import PCA

class scMMT_API(object):
    def __init__(self, gene_trainsets, protein_trainsets=[], gene_test = None, gene_list = [], select_hvg = True, train_batchkeys = None, test_batchkey = None, type_key = None, cell_normalize = True, log_normalize = True, gene_normalize = True, min_cells = 30, min_genes = 200, n_svd = 300, n_fa=64, n_hvg=1000,dataset_batch = True,data_dir = None,data_load=False,add_meta=[],batch_size = 128, val_split = "by_test",val_frac=0.1, use_gpu = True,seed=5,log_weight=None):                
        
        if use_gpu:
            print("Searching for GPU")
            
            if is_available():
                print("GPU detected, using GPU")
                self.device = 'cuda'
                
            else:
                print("GPU not detected, falling back to CPU")
                self.device = 'cpu'
                
        else:
            print("Using CPU")
            self.device = 'cpu'
            
        preprocess_args = (gene_trainsets, protein_trainsets, gene_test, train_batchkeys, test_batchkey, type_key,
                     gene_list, select_hvg, cell_normalize, log_normalize, gene_normalize, min_cells,                                                  min_genes,n_svd, n_fa, n_hvg,dataset_batch,data_dir,data_load,seed)

        genes, proteins, genes_test, bools, train_keys, categories = preprocess(*preprocess_args)
        
        if len(add_meta)>0:   
            genes.obsm['result'] = np.concatenate([genes.obsm['result'],genes.obs[add_meta]],axis=1)
            genes_test.obsm['result'] = np.concatenate([genes_test.obsm['result'],genes_test.obs[add_meta]],axis=1)        
        
        self.proteins = proteins
        self.gene = genes.obsm['result']
        self.train_cells = genes.obs.copy()
        self.type_key = type_key
        self.categories = categories
        
        if log_weight == "no_log":
            label = genes.obs[type_key].replace(categories).values
            self.weight = tensor(compute_class_weight(class_weight='balanced', classes=list(categories.values()), y=label),dtype=torch.float32, device = self.device)
        elif log_weight is not None:
            label = genes.obs[type_key].replace(categories).values
            self.weight = tensor(np.log(compute_class_weight(class_weight='balanced', classes=list(categories.values()), y=label)+log_weight) ,dtype=torch.float32, device = self.device)
        else:
            self.weight = None
        
        if genes_test is not None:
            self.test_cells = genes_test.obs.copy()
        else:
            self.test_cells = None
            
        if categories is not None:
            celltypes = proteins.obs[type_key]
        else:
            celltypes = None
        
        if val_split == "by_test":
            print("PCA")
            pca = PCA(n_components = 100)
            train = pca.fit_transform(genes.X)
            test = pca.transform(genes_test.X) 
            print("KNN")
            neigh = NearestNeighbors(n_neighbors=1,metric="cosine")
            neigh.fit(train)
            temp = neigh.kneighbors(test,return_distance=False)
            temp = list(set(temp.reshape(1,-1)[0]))
            val_split = temp
        
        dataloaders = build_dataloaders(genes, proteins, genes_test, bools,train_keys,val_split,val_frac,batch_size,self.device,celltypes,categories)
        self.dataloaders = {key: loader for key, loader in zip(['train', 'val', 'impute', 'test'], dataloaders)}
        
    def train(self,n_epochs = 10000, ES_max = 12, decay_max = 6, h_size=512,drop_rate=0.25,n_layer=4,label_smoothing=0.01, 
              decay_step = 0.1, lr = 10**(-3), weights_dir = None,load = False):
            
        protein_mloss = mse_loss()
        
        if self.categories is not None:
            type_loss = cross_entropy
        else:
            type_loss = no_loss(self.device)
        
        p_mod1, p_mod2 = self.gene.shape[1], self.proteins.obsm["result"].shape[1]
        model_params = {'p_mod1': p_mod1, 'p_mod2': p_mod2, 'loss1': type_loss, 'loss2': protein_mloss, 'categories': self.categories, 'weight':self.weight, "h_size":h_size, "drop_rate":drop_rate, "n_layer":n_layer, "label_smoothing":label_smoothing}

        self.model = scMMT_Model(**model_params)
        self.model.to(self.device)
        
#         build_dir(weights_dir)
        path = weights_dir
        if load and isfile(path):
            self.model.load_state_dict(torch_load(path))
            
        else:
            train_params = (self.dataloaders['train'], self.dataloaders['val'], n_epochs, ES_max, decay_max, decay_step, lr, path,self.device)
        
            self.model.train_backprop(*train_params)
            self.model.load_state_dict(torch_load(path))
#             save(self.model.state_dict(), path_new)
        
    def impute(self,):
        return self.model.impute(self.dataloaders['impute'], self.proteins)
        
    def predict(self,):
        assert self.test_cells is not None
        return self.model.predict(self.dataloaders['test'], self.proteins, self.test_cells)
    
    def embed(self,):
        if self.test_cells is not None:
            loaders = self.dataloaders['impute'], self.dataloaders['test']
        else:
            loaders = self.dataloaders['impute'], None
        return self.model.embed(*loaders, self.train_cells, self.test_cells)



from os.path import split, basename, isdir
from os import mkdir
from anndata import AnnData
from scipy.sparse import issparse, csr_matrix,csc_matrix
format_loss = lambda loss, nbatch: round(loss/nbatch, 2)
import numpy as np
def build_dir(dir_path):
    """ This function builds a directory if it does not exist.
    
    
    Arguments:
    ------------------------------------------------------------------
    - dir_path: `str`, The directory to build. E.g. if dir_path = 'folder1/folder2/folder3', then this function will creates directory if folder1 if it does not already exist. Then it creates folder1/folder2 if folder2 does not exist in folder1. Then it creates folder1/folder2/folder3 if folder3 does not exist in folder2.
    """
    
    subdirs = [dir_path]
    substring = dir_path

    while substring != '':
        splt_dir = split(substring)
        substring = splt_dir[0]
        subdirs.append(substring)
        
    subdirs.pop()
    subdirs = [x for x in subdirs if basename(x) != '..']
    
    for dir_ in subdirs[::-1]:
        if not isdir(dir_):
            mkdir(dir_)

            
def clr(adata=AnnData, inplace= True, axis = 0):
    """
    Apply the centered log ratio (CLR) transformation
    to normalize counts in adata.X.
    Args:
        data: AnnData object with protein expression counts.
        inplace: Whether to update adata.X inplace.
        axis: Axis across which CLR is performed.
    """

    if axis not in [0, 1]:
        raise ValueError("Invalid value for `axis` provided. Admissible options are `0` and `1`.")

    if not inplace:
        adata = adata.copy()

    if issparse(adata.X) and axis == 0 and not isinstance(adata.X, csc_matrix):
        warn("adata.X is sparse but not in CSC format. Converting to CSC.")
        x = csc_matrix(adata.X)
    elif issparse(adata.X) and axis == 1 and not isinstance(adata.X, csr_matrix):
        warn("adata.X is sparse but not in CSR format. Converting to CSR.")
        x = csr_matrix(adata.X)
    else:
        x = adata.X
        
    if issparse(x):
        
        x.data /= np.repeat(
            np.exp(np.log1p(x).sum(axis=axis).A / x.shape[axis]), x.getnnz(axis=axis)
        )
        np.log1p(x.data, out=x.data)
    else:
        np.log1p(
            x / np.exp(np.log1p(x).sum(axis=axis, keepdims=True) / x.shape[axis]),
            out=x,
        )

    adata.X = x

    return None if inplace else adata 


def make_dense(anndata):
    if issparse(anndata.X):
        tmp = anndata.X.copy()
        anndata.X = tmp.copy().toarray()


from torch import tensor, long
import numpy as np
def tensor_loader(*adatas, protein_boolean, sampler, device, celltypes = None, categories = None):
    if (celltypes is not None) and (categories is not None):
        return tensor_loader_trihead(*adatas, protein_boolean = protein_boolean, sampler = sampler, device = device, celltypes = celltypes, categories = categories)
    
    return tensor_loader_dualhead(*adatas, protein_boolean = protein_boolean, sampler = sampler, device = device)
    
class tensor_loader_dualhead:
    def __init__(self, *adatas, protein_boolean, sampler, device):
        assert all([adata.shape[0] == adatas[0].shape[0] for adata in adatas])
        
        self.arrs = [adata.obsm['result'] for adata in adatas]
#         print(arrs.shape)
        self.bools = protein_boolean
        
        self.sampler = sampler
        self.device = device
        
    def __iter__(self):
        for idxs, bool_set in self.sampler:
            arrays = [arr[idxs] for arr in self.arrs]
            arrays = [tensor(arr, device = self.device) for arr in arrays]
            arrays.append(tensor(self.bools[bool_set], device = self.device))
            arrays.append(None)
            
            yield arrays
    
    def __len__(self):
        return len(self.sampler)
    
class tensor_loader_trihead:
    def __init__(self, *adatas, protein_boolean, sampler, device, celltypes, categories):
        assert all([adata.shape[0] == adatas[0].shape[0] for adata in adatas])
        
        self.arrs =  [adata.obsm['result'] for adata in adatas]
        self.celltypes, self.categories = celltypes, categories
        self.bools = protein_boolean
        
        self.sampler = sampler
        self.device = device
        
    def __iter__(self):
        for idxs, bool_set in self.sampler:
            arrays = [arr[idxs] for arr in self.arrs]
            arrays = [tensor(arr, device = self.device) for arr in arrays]
            arrays.append(tensor(self.bools[bool_set], device = self.device))
            arrays.append(tensor([self.categories[cat] for cat in self.celltypes[idxs]], device = self.device, dtype = long))
                        
            yield arrays
    
    def __len__(self):
        return len(self.sampler)
    
    
class tensor_loader_basic:
    def __init__(self, adata, batch_size, device):
        self.arr = adata.obsm['result'].copy()#np.concatenate((adata.X,adata.obsm['X_svd'],adata.obsm['X_fa']),axis=1).copy()
        self.batch_size = batch_size
        self.device = device
        
    def __iter__(self):
        batch = []
        
        for idx in range(len(self.arr)):
            batch.append(idx)
            
            if len(batch) == self.batch_size:
                yield tensor(self.arr[batch], device = self.device)
                batch = []
                
        if batch:
            yield tensor(self.arr[batch], device = self.device)
            
    def __len__(self):
        return len(self.arr)


from torch import tensor, abs as torch_abs, logical_not, log, clamp
from torch.nn import Softplus
import torch

class no_loss(object):
    def __init__(self, device):
        self.device = device
    
    def __call__(self, outputs, target):
        return tensor(0., device = self.device)
    
    
class mse_loss(object):
    def __init__(self, reduce = True):
        self.reduce = reduce
        
    def __call__(self, yhat, y, bools=None):
        SEs = (yhat - y)**2
        
        if self.reduce:
            if bools != None:
                loss = SEs.mean() * bools
                return loss.mean()
            return SEs.mean()    
        return SEs


from numpy import arange, setdiff1d
from numpy.random import choice

class batchSampler:
    def __init__(self, indices, train_keys, bsize, shuffle = False):
        self.indices = indices
        self.train_keys = train_keys
        self.bsize = bsize
        self.shuffle = shuffle
        
    def __iter__(self):
        if self.shuffle:
            indices = choice(self.indices, size = len(self.indices), replace = False)
        else:
            indices = self.indices
        
        minibatch_idx = []
        bool_idx = []
        
        for idx in indices:
            minibatch_idx.append(idx)
            bool_idx.append(sum([int(idx >= x) for x in self.train_keys]))
            
            if len(minibatch_idx) >= self.bsize:
                yield minibatch_idx, bool_idx
                minibatch_idx, bool_idx = [], []
                
        if minibatch_idx:
            yield minibatch_idx, bool_idx
        
    def __len__(self):
        return len(self.indices)
    
def build_trainSamplers(adata, n_train, bsize = 128,val_split = "by_test",val_frac=0.1):
    num_val = round(val_frac * len(adata))
    if val_split == "by_test":
        val_idx = val_split
        if len(val_idx)>num_val:
            val_idx = choice(val_idx, num_val, replace = False)
    else:
        assert num_val >= 1

        idx = arange(len(adata))
        val_idx = choice(idx, num_val, replace = False)

    train_indices, val_indices = setdiff1d(idx, val_idx).tolist(), val_idx.tolist()

    train_sampler = batchSampler(train_indices, n_train, bsize, shuffle = True)
    val_sampler = batchSampler(val_indices, n_train, bsize)
    
    return train_sampler, val_sampler

def build_testSampler(adata, train_keys, bsize = 128):
    indices = list(range(len(adata)))
    
    return batchSampler(indices, train_keys, bsize)


from .Samplers import build_trainSamplers, build_testSampler
from .DataLoader import tensor_loader, tensor_loader_basic

def build_dataloaders(gene_train, protein_train, gene_test, bools, train_keys, val_split, val_frac, batch_size, device, celltypes = None, categories = None):
    sampler_train, sampler_val = build_trainSamplers(gene_train, train_keys, batch_size, val_split, val_frac)
    indexer_test = build_testSampler(gene_train, train_keys, batch_size)

    DS_train = tensor_loader(gene_train, protein_train, protein_boolean = bools, sampler = sampler_train, device = device, celltypes = celltypes, categories = categories)
    DS_val = tensor_loader(gene_train, protein_train, protein_boolean = bools, sampler = sampler_val, device = device, celltypes = celltypes, categories = categories)
    DS_impute = tensor_loader(gene_train, protein_boolean = bools, sampler = indexer_test, device = device, celltypes = celltypes, categories = categories)
    
    if gene_test is not None:
        DS_test = tensor_loader_basic(gene_test, batch_size, device = device)
    else:
        DS_test = None
    
    return DS_train, DS_val, DS_impute, DS_test


from sklearnex import patch_sklearn, unpatch_sklearn
patch_sklearn()
from tqdm import tqdm
from time import sleep
import numpy as np
from numpy import intersect1d, setdiff1d, quantile, unique, asarray, zeros
from numpy.random import choice, seed
from .Utils import make_dense
import pickle
from anndata import AnnData
import scanpy as sc
import pandas as pd
import torch
from sklearn.decomposition import TruncatedSVD, FactorAnalysis

sc.settings.verbosity = 0

def preprocess(gene_trainsets, protein_trainsets=[], gene_test = None, train_batchkeys = None, test_batchkey = None, type_key = None, gene_list = [], select_hvg = True, cell_normalize = True, log_normalize = True, gene_normalize = True, min_cells = 1, min_genes = 1,n_svd = 300, n_fa=180, n_hvg=550,dataset_batch = True,data_dir="data.pkl",data_load=False,seed=5):
    
    assert type(gene_trainsets) == list
    assert type(protein_trainsets) == list
    assert all([sum(x.obs.index == y.obs.index)/len(x) == 1. for x, y in zip(gene_trainsets, protein_trainsets)])
    
    if protein_trainsets==[]:
        for tmp in gene_trainsets:
            protein_tmp = AnnData(np.zeros(shape = (len(tmp.obs), 0)),obs=tmp.obs,dtype=np.int32)
            protein_trainsets.append(protein_tmp)    
            
    seed = seed
    torch.manual_seed(seed) 
    torch.cuda.manual_seed(seed) 
    torch.cuda.manual_seed_all(seed) 
    np.random.seed(seed) 
    
    if data_dir is not None and data_load:
        #print(data_dir,data_load)
        data = pickle.load(open(data_dir, 'rb'))
        gene_train, protein_train, gene_test, bools, train_keys, categories = data
        return gene_train, protein_train, gene_test, bools, train_keys, categories
    
    if type_key:
        assert all([type_key in x.obs.columns for x in protein_trainsets])
        categories, i = {}, 0
        
        for dataset in protein_trainsets:
            for celltype in dataset.obs[type_key]:
                if celltype not in categories:
                    categories[celltype] = i
                    i += 1
     
    else:
        categories = None
            
    if train_batchkeys is not None:
        assert type(train_batchkeys) == list
        for i in range(len(gene_trainsets)):
            key = train_batchkeys[i]
            gene_trainsets[i].obs['batch'] = ['DS-' + str(i + 1) + ' ' + x for x in gene_trainsets[i].obs[key]]
            protein_trainsets[i].obs['batch'] = ['DS-' + str(i + 1) + ' ' + x for x in protein_trainsets[i].obs[key]]
    else:
        for i in range(len(gene_trainsets)):
            gene_trainsets[i].obs['batch'] = 'DS-' + str(i + 1)
            protein_trainsets[i].obs['batch'] = 'DS-' + str(i + 1)
        
    if gene_test is not None:
        if test_batchkey is not None:
            gene_test.obs['batch'] = ['DS-Test ' + x for x in gene_test.obs[test_batchkey]]
        else:
            gene_test.obs['batch'] = 'DS-Test'
                    
        
    gene_set = set(gene_list)
    
    if min_genes:
        print("\nQC Filtering Training Cells")
        
        for i in range(len(gene_trainsets)):
            cell_filter = (gene_trainsets[i].X > 10**(-8)).sum(axis = 1) >= min_genes
            gene_trainsets[i] = gene_trainsets[i][cell_filter].copy()
            protein_trainsets[i] = protein_trainsets[i][cell_filter].copy()
        
        if gene_test is not None:
            print("QC Filtering Testing Cells")
            
            cell_filter = (gene_test.X > 10**(-8)).sum(axis = 1) >= min_genes
            gene_test = gene_test[cell_filter].copy()
        
    if min_cells:
        print("\nQC Filtering Training Genes")
        
        for i in range(len(gene_trainsets)):
            bools = (gene_trainsets[i].X > 10**(-8)).sum(axis = 0) >= min_cells            
            genes = gene_trainsets[i].var.index[bools]
            genes = asarray(genes).reshape((-1,))
            features = set(genes)
            features.update(gene_set)
            features = list(features)
            features.sort()

            gene_trainsets[i] = gene_trainsets[i][:, features].copy()

        if gene_test is not None:
            print("QC Filtering Testing Genes")

            bools = (gene_test.X > 10**(-8)).sum(axis = 0) >= min_cells
            genes = gene_test.var.index[bools]
            genes = asarray(genes).reshape((-1,))
            features = set(genes)
            features.update(gene_set)
            features = list(features)
            features.sort()

            gene_test = gene_test[:, features].copy()
    for i in range(len(gene_trainsets)):
        gene_trainsets[i].layers["raw"] = gene_trainsets[i].X.copy()
        protein_trainsets[i].layers["raw"] = protein_trainsets[i].X.copy()
    if gene_test is not None:
        gene_test.layers["raw"] = gene_test.X.copy()
            
    if cell_normalize:
        print("\nNormalizing Training Cells")
        
        [sc.pp.normalize_total(x) for x in gene_trainsets]

        if gene_test is not None:
            print("Normalizing Testing Cells")
            sc.pp.normalize_total(gene_test, key_added = "scale_factor")

    if log_normalize:
        print("\nLog-Normalizing Training Data")
        
        [sc.pp.log1p(x) for x in gene_trainsets]
        
        if gene_test is not None:
            print("Log-Normalizing Testing Data")
            sc.pp.log1p(gene_test)
    
    gene_train = gene_trainsets[0]
    gene_train.obs['Dataset'] = 'Dataset 1'
    protein_trainsets[0].obs['Dataset'] = 'Dataset 1'
    for i in range(1, len(gene_trainsets)):
        gene_trainsets[i].obs['Dataset'] = 'Dataset ' + str(i + 1)
        protein_trainsets[i].obs['Dataset'] = 'Dataset ' + str(i + 1)
        gene_train = gene_train.concatenate(gene_trainsets[i], join = 'inner', batch_key = None)
    
    if gene_test is not None:
        genes = intersect1d(gene_train.var.index, gene_test.var.index)
        gene_train = gene_train[:, genes].copy()
        gene_test = gene_test[:, genes].copy()
              
    make_dense(gene_train)
    [make_dense(x) for x in protein_trainsets]
    if gene_test is not None:
        make_dense(gene_test)

    #demension
    tsvd = TruncatedSVD(n_components = n_svd)
    FA = FactorAnalysis(n_components = n_fa)
    if gene_test is not None:
        temp = np.concatenate((np.array(gene_train.X),np.array(gene_test.X)),axis=0)
        if dataset_batch: 
            tmp = AnnData(temp)
            gene_test.obs["dataset"] = "test"
            gene_train.obs["dataset"] = "train"
            tmp.obs = pd.concat([gene_train.obs,gene_test.obs],axis=0)
            print("\ncombat")
            sc.pp.combat(tmp, key='dataset')
            temp = tmp.X.copy()
                 
        print("\nTSVD...")
        temp_svd = tsvd.fit_transform(temp)
        train_col = gene_train.X.shape[0]
        gene_train.obsm['X_svd'] = temp_svd[:train_col]
        gene_test.obsm['X_svd'] = temp_svd[train_col:]
        
#         gene_train.X = temp[:train_col]
#         gene_test.X = temp[train_col:]
        
        print("\nFA ...")
        temp_fa = FA.fit_transform(temp)
        gene_train.obsm['X_fa'] = temp_fa[:train_col]
        gene_test.obsm['X_fa'] = temp_fa[train_col:]
    else:
        print("\nTSVD Train")
        gene_train.obsm['X_svd'] = tsvd.fit_transform(gene_train.X)
        print("\nFA train")
        gene_train.obsm['X_fa'] = FA.fit_transform(gene_train.X)
       
    
    if select_hvg:
        print("\nFinding HVGs")
        
        if gene_test is not None:
            tmp = gene_train.concatenate(gene_test, batch_key = None).copy()
        else:
            tmp = gene_train.copy()
        
        if not cell_normalize or not log_normalize:
            print("Warning, highly variable gene selection may not be accurate if expression is not cell normalized and log normalized")
            
        if len(tmp) > 10**5:
            idx = choice(range(len(tmp)), 10**5, False)
            tmp = tmp[idx].copy()
            
        sc.pp.highly_variable_genes(tmp, min_mean = 0.0125, max_mean = 3, min_disp = 0.5, 
                              n_bins = 20, subset = False, batch_key = 'batch', n_top_genes = n_hvg)
        hvgs = tmp.var.index[tmp.var['highly_variable']].copy()
        tmp = None
        
        gene_set.update(set(hvgs))
        gene_set = list(gene_set)
        gene_set.sort()
        gene_train = gene_train[:, gene_set].copy()
        if gene_test is not None:
            gene_test = gene_test[:, gene_set].copy()
    
    if gene_normalize:
        patients = unique(gene_train.obs['batch'].values)

        print("\nNormalizing Gene Training Data by Batch")
        sleep(1)

        for patient in patients:
            indices = [x == patient for x in gene_train.obs['batch']]
            sub_adata = gene_train[indices].copy()
            sc.pp.scale(sub_adata)

            gene_train[indices] = sub_adata.X.copy()
                
        if gene_test is not None:
            patients = unique(gene_test.obs['batch'].values)

            print("\nNormalizing Gene Testing Data by Batch")
            
            
            for patient in patients:
                indices = [x == patient for x in gene_test.obs['batch']]
                sub_adata = gene_test[indices].copy()
                sc.pp.scale(sub_adata)

                gene_test[indices] = sub_adata.X.copy()

        train_keys, curr_break, proteins = [], len(protein_trainsets[0]), [set(protein_trainsets[0].var.index)]
        protein_train = protein_trainsets[0].copy()
        for i in range(1, len(protein_trainsets)):
            protein_train = protein_train.concatenate(protein_trainsets[i], join = 'outer', fill_value = 0., 
                                                      batch_key = None)
            
            proteins.append(set(protein_trainsets[i].var.index))
            train_keys.append(curr_break)
            curr_break += len(protein_trainsets[i])
        
        bools = asarray([[int(x in prot_set) for x in protein_train.var.index] for prot_set in proteins])
        
        for i in range(len(bools)):
            protein_train.var['Dataset ' + str(i + 1)] = [bool(x) for x in bools[i]]
        
    gene_train.obsm['result'] = np.concatenate([gene_train.obsm['X_svd'],gene_train.obsm['X_fa'],gene_train.X],axis=1)
    if gene_test is not None:
        gene_test.obsm['result'] = np.concatenate([gene_test.obsm['X_svd'],gene_test.obsm['X_fa'] ,gene_test.X],axis=1)
    
    protein_train.obsm['result'] = protein_train.X
    
    data = (gene_train, protein_train, gene_test, bools, train_keys, categories)
    pickle.dump(data,open(data_dir, 'wb'))
    return gene_train, protein_train, gene_test, bools, train_keys, categories
        


from torch import cat
from torch.nn import Module, Linear, BatchNorm1d, PReLU, Dropout,SELU,Tanh,GELU, RNNCell
import torch

class Input_Block(Module):
    def __init__(self, in_units, out_units, dropout_inrate, dropout_outrate):
        super(Input_Block, self).__init__()
        self.bnorm_in = BatchNorm1d(in_units)
        self.dropout_in = Dropout(dropout_inrate)
        self.dense = Linear(in_units, out_units)
        self.act = SELU()
            
    def forward(self, x_new):
        x_new = self.bnorm_in(x_new)
        x_new = self.dropout_in(x_new)
        x = self.dense(x_new)
        h = self.act(x)
        return h
    
class Resnet(Module):
    def __init__(self, hidden_units, dropout_rate=0.1):
        super(Resnet, self).__init__()
        self.Dropout = Dropout(dropout_rate)
        self.dense = Linear(hidden_units, hidden_units)
        self.act =  SELU()
        
    def forward(self, x):
        h = self.Dropout(x)
        h = self.dense(h)
        h = self.act(h)
        return x+h

class Resnet_last(Module):
    def __init__(self, hidden_units, dropout_rate=0.1):
        super(Resnet_last, self).__init__()
        self.Dropout = Dropout(dropout_rate)
        self.dense = Linear(hidden_units, hidden_units)
        self.act =  SELU()
        
    def forward(self, x):
        x = self.Dropout(x)
        x = self.dense(x)
        x = self.act(x)
        return x


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="scMMT",
    version="1.0.1",
    author="Songqi Zhou",
    author_email="zhousongqi21@gmail.com",
    description="A package for cell annotation, protein prediction, and low dimensional embedding representation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SongqiZhou/scMMT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=['torch>=2.0.0',  'scanpy>=1.9.3', 'scikit-learn>=1.2.2','scikit-learn-intelex>=2023.1.1','pandas>=2.0.1', 'numpy>=1.24.3', 'scipy>=1.10.1', 'tqdm>=4.65.0', 'anndata>=0.9.1'],
    python_requires=">=3.10",
)


from math import log, exp
from numpy import inf, zeros, zeros_like as np_zeros_like, arange, asarray, empty
from pandas import concat
from anndata import AnnData
import torch
import numpy as np

from torch import cat, no_grad, randn, zeros_like, zeros as torch_zeros, ones, argmax
from torch.nn import Module, Linear, Sequential, RNNCell, Softplus, Parameter, Softmax, SELU,Tanh,L1Loss
from torch.optim import Adam
from torch.optim.lr_scheduler import StepLR

from .Layers import Input_Block, Resnet,Resnet_last


from torch.nn import Module, Linear, BatchNorm1d, PReLU, Dropout, GRU ,AlphaDropout, Tanh
class scMMT_Model(Module):
    def __init__(self, p_mod1, p_mod2, loss1, loss2, categories, weight,h_size, drop_rate,n_layer,label_smoothing):
        super(scMMT_Model, self).__init__()
        self.p_mod2 = p_mod2
        h_size, drop_rate = 512, 0.25
        
        self.input_block = Input_Block(p_mod1, h_size, drop_rate, drop_rate)       
        self.resnet = Sequential(*[Resnet(h_size) for i in range(n_layer)])        
        self.resnet_last = Resnet_last(h_size)
        MSE_output = Sequential(Linear(h_size, p_mod2))
        
        self.mod2_out = MSE_output
                
        if categories is not None:
            self.celltype_out = Sequential(Linear(h_size, len(categories)))
            self.forward = self.forward_transfer
            self.categories_arr = empty((len(categories), ), dtype = 'object')
            for cat in categories:
                self.categories_arr[categories[cat]] = cat
        else:
            self.forward = self.forward_simple
            self.categories_arr = None
            
        self.loss1, self.loss2 = loss1, loss2
        self.label_smoothing = label_smoothing
        self.weight = weight
        
    def forward_transfer(self, x):
        
        x = x.to(torch.float32)
        x = self.input_block(x)
        h = self.resnet(x)
        h = self.resnet_last(h)
        return {'celltypes': self.celltype_out(h), 'modality 2': self.mod2_out(h), 'embedding': h}
 
    def forward_simple(self, x):
        
        x = x.to(torch.float32)
        x = self.input_block(x)
        h = self.resnet(x)
        h = self.resnet_last(h)
        return {'celltypes': None, 'modality 2': self.mod2_out(h), 'embedding': h}
    
    def train_backprop(self, train_loader, val_loader,
                 n_epoch = 10000, ES_max = 30, decay_max = 10, decay_step = 0.1, lr = 10**(-3), path=None, device="cpu"):
        alph = torch.Tensor([0.15]).to(device)
        ES_max = torch.Tensor([ES_max]).to(device)
        if self.categories_arr is None:
            optimizer = Adam(self.parameters(), lr = lr)
            scheduler = StepLR(optimizer, step_size = 1, gamma = decay_step)
            patience = torch.zeros(1).to(device)
            bestloss = torch.Tensor([np.inf]).to(device)
            get_correct = lambda x: 0
            for epoch in range(n_epoch):
                self.train()
                for batch, inputs in enumerate(train_loader):
                    optimizer.zero_grad()
                    mod1, mod2, protein_bools, celltypes = inputs    
                    outputs = self(mod1)
                    mod2_loss = self.loss2(outputs['modality 2'], mod2, protein_bools)
                    loss = mod2_loss
                    loss.backward()
                    optimizer.step()
                with no_grad():
                    running_loss, rtype_acc = torch.zeros(1).to(device), torch.zeros(1).to(device)
                    self.eval()
                    for batch, inputs in enumerate(val_loader):
                        mod1, mod2, protein_bools, celltypes = inputs
                        outputs = self(mod1)
                        n_correct = get_correct(outputs)
                        mod2_loss = self.loss2(outputs['modality 2'], mod2, protein_bools)
                        rtype_acc += n_correct
                        running_loss += mod2_loss.item() * len(mod2)
                    print(f"Epoch {epoch} prediction loss = {running_loss.item()/len(val_loader):.3f}")
                    patience += 1
                    if bestloss > running_loss:
                        bestloss, patience = running_loss, 0
                        torch.save(self.state_dict(),path)
                    if (patience + 1) % decay_max == 0:
                        scheduler.step()
                        print(f"Decaying loss to {optimizer.param_groups[0]['lr']}")
                    if (patience + 1) > ES_max:
                        break
      
        else:
            get_correct = lambda outputs: (argmax(outputs['celltypes'], axis = 1) == celltypes).sum()        
            Weightloss1 = torch.FloatTensor([1]).to(device).requires_grad_(True)
            Weightloss2 = torch.FloatTensor([1]).to(device).requires_grad_(True)
            params = [Weightloss1, Weightloss2]
            optimizer1 = torch.optim.Adam(self.parameters(), lr=lr)
            optimizer2 = torch.optim.Adam(params, lr=lr)
            scheduler1 = StepLR(optimizer1, step_size = 1, gamma = decay_step)
            scheduler2 = StepLR(optimizer2, step_size = 1, gamma = decay_step)      
            patience = torch.zeros(1).to(device)
            bestloss = torch.Tensor([np.inf]).to(device)
            bestloss2 = torch.Tensor([np.inf]).to(device)
            bestacc = torch.zeros(1).to(device)
            Gradloss = L1Loss()        
            for epoch in range(n_epoch):
                self.train()
                for batch, inputs in enumerate(train_loader):
                    mod1, mod2, protein_bools, celltypes = inputs    
                    outputs = self(mod1)
                    mod1_loss = params[0]*self.loss1(outputs['celltypes'], celltypes, weight = self.weight,label_smoothing=self.label_smoothing)
                    mod2_loss = params[1]*self.loss2(outputs['modality 2'], mod2, protein_bools)
                    loss = torch.div(mod1_loss+mod2_loss, 2)
                    if epoch == 0:
                        l01 = mod1_loss.data  
                        l02 = mod2_loss.data
                    optimizer1.zero_grad()
                    loss.backward(retain_graph=True) 
                    param = list(self.parameters())
                    G1R = torch.autograd.grad(mod1_loss, param[0], retain_graph=True, create_graph=True)
                    G1 = torch.norm(G1R[0], 2)
                    G2R = torch.autograd.grad(mod2_loss, param[0], retain_graph=True, create_graph=True)
                    G2 = torch.norm(G2R[0], 2)
                    G_avg = torch.div(G1+G2, 2)
                    lhat1 = torch.div(mod1_loss,l01)
                    lhat2 = torch.div(mod2_loss,l02)
                    lhat_avg = torch.div(lhat1+lhat2, 2)
                    inv_rate1 = torch.div(lhat1,lhat_avg)
                    inv_rate2 = torch.div(lhat2,lhat_avg)
                    C1 = G_avg*(inv_rate1)**alph
                    C2 = G_avg*(inv_rate2)**alph
                    C1 = C1.detach()
                    C2 = C2.detach()
                    optimizer2.zero_grad()
                    Lgrad = Gradloss(G1, C1)+Gradloss(G2, C2)
                    Lgrad.backward()
                    optimizer2.step()
                    optimizer1.step()
                    coef = 2/(Weightloss1+Weightloss2)
                    params = [coef*Weightloss1, coef*Weightloss2]
                with no_grad():
                    running_loss,rtype_acc =torch.zeros(1).to(device),torch.zeros(1).to(device)
                    self.eval()
                    for batch, inputs in enumerate(val_loader):
                        mod1, mod2, protein_bools, celltypes = inputs
                        outputs = self(mod1)
                        n_correct = get_correct(outputs)
                        mod4_loss = self.loss2(outputs['modality 2'], mod2, protein_bools)
                        rtype_acc += n_correct
                        running_loss += mod4_loss.item() * len(mod2)
                        
                    print(f"Epoch {epoch} validation accuracy = {rtype_acc.item()/len(val_loader):.3f},mseloss={running_loss.item()/len(val_loader):.3f}")
                    patience += 1                              
                    if ((bestloss / (running_loss/len(val_loader))).item() +\
                        ((rtype_acc/len(val_loader))/bestacc)).item()>2:
                        bestloss,bestacc, patience = running_loss/len(val_loader), rtype_acc/len(val_loader), torch.zeros(1).to(device)
                        torch.save(self.state_dict(),path)
                    if self.p_mod2 == 0:
                        if ((rtype_acc/len(val_loader))/bestacc).item()>1:
                            bestacc, patience = rtype_acc/len(val_loader), torch.zeros(1).to(device)
                            torch.save(self.state_dict(),path)
                        
                    if (patience + 1) % decay_max == 0:
                        scheduler1.step()
                        scheduler2.step()
                        print(f"Decaying loss to {optimizer1.param_groups[0]['lr']}")
                    if (patience + 1) > ES_max:
                        break
    
    def impute(self, impute_loader, proteins):
        imputed_test = proteins.copy()
        self.eval()
        start = 0
        for mod1, bools, celltypes in impute_loader:
            end = start + mod1.shape[0]
            with no_grad():
                outputs = self(mod1)
            mod2_impute = outputs['modality 2']
            imputed_test.X[start:end] = self.fill_predicted(imputed_test.X[start:end], mod2_impute, bools)
            start = end
        return imputed_test
    
    def embed(self, impute_loader, test_loader, cells_train, cells_test):
        if cells_test is not None:
            embedding = AnnData(zeros(shape = (len(cells_train) + len(cells_test), 512)))
            embedding.obs = concat((cells_train, cells_test), join = 'inner')
        else:
            embedding = AnnData(zeros(shape = (len(cells_train), 512)))
            embedding.obs = cells_train
        self.eval()
        start = 0
        for mod1, bools, celltypes in impute_loader:
            end = start + mod1.shape[0]
            outputs = self(mod1)
            embedding[start:end] = outputs['embedding'].detach().cpu().numpy()
            start = end
        if cells_test is not None:
            for mod1 in test_loader:
                end = start + mod1.shape[0]
                outputs = self(mod1)
                embedding[start:end] = outputs['embedding'].detach().cpu().numpy()
                start = end
        return embedding
                    
    def fill_predicted(self, array, predicted, bools):
        bools = bools.cpu().numpy()
        return (1. - bools) * predicted.cpu().numpy() + array
    
    def predict(self, test_loader, proteins, cells):
        imputed_test = AnnData(zeros(shape = (len(cells), self.p_mod2)),dtype=np.float32)
        imputed_test.obs = cells
        imputed_test.var.index = proteins.var.index
        if self.categories_arr is not None:
            celltypes = ['None'] * len(cells)
        self.eval()
        start = 0
        for mod1 in test_loader:
            end = start + mod1.shape[0]
            with no_grad():
                outputs = self(mod1)
            if self.categories_arr is not None:
                predicted_types = argmax(outputs['celltypes'], axis = 1).cpu().numpy()
                celltypes[start:end] = self.categories_arr[predicted_types].tolist()
            mod2_impute = outputs['modality 2']
            imputed_test[start:end] = mod2_impute.cpu().numpy()
            start = end
        if self.categories_arr is not None:
            imputed_test.obs['transfered cell labels'] = celltypes
        return imputed_test